#!/usr/bin/perl -w

###############################################################################
# send_nrdp.pl - Perl implementation of an NRDP client.
#
# FEATURES:
# - Connects to NRDP CGI (http port 80) and submits passive check results.
# - Syntaxually compatible with send_nsca, and designed as a 'drop-in'
#   replacement for send_nsca.	Requires some additions to the send_nsca
#   configuration file to fully function.
# - Optionally caches results that fail to send to attempt to re-send during
#   the next execution.  So when a commit is ocurring, or an outtage occurs
#   alarms aren't lost.
# - Optionally writes execution to a log file for debugging, or confirmation
#   of execution.
# - Optionally submits results to a second nrdp host.  Useful in distributed
#   configurations where there is a primary/backup pair.  Alarms can be sent
#   to both, and whoever is configured to send notifications will notify
#   about the alarm.
# - Based off 'send_nrdp.php' that comes with the nrdp project.  We don't have
#   the php binary installed on our monitored Unix hosts, so we wrote this.
# - Minimal perl module requirement.  Should run on native perl included with
#   most modern Unixes (Needs IO::Socket, and Getopt::Long).
# - Allows for basic http authentication with a user and password.
# - No support for SSL (would require more libraries, and this is meant to be
#   light!).
# - Tested on RHEL 4/5, Solaris 10/10x86, and AIX 5.3/6.1.
# - Used in our production environment (5k hosts, 45k services, distributed
#   nagios) for about 2 months now.
#
# HISTORY:
# v1.0 - Adam McLean - Initial release.
###############################################################################

use strict;
use IO::Socket;
use Getopt::Long;
require 5;

my $version="1.0";

# At a minimum we'd expect 2 arguments (-H and the hostname).  If we don't get them
# then we can print help and exit 0.
if(scalar(@ARGV) < 2) {
	&help(0);
}

# $gconf will be used globally by all subs to hold our configuration elements.
my $gconf={};
############################# $gconf data format #################################
# {host} - Primary hostname to send alarms to (from -H commandline)
# {delim} - Delimiter to use for the alarms to STDIN (from -d commandline)
# {token} - Token to send to nrdp for events (from token= in config file or --token on command line)
# {timeout} - Amount of time before we time out (from -to commandline)
# {old_times} - Whether old timestamps were specified on the commandline (-od).  send_nrdp doesn't appear to support
#		adding a timestamp to the XML forward, it's here for future use though.
# {url} - URL to find nrdp listening on (from --url on commandline or url= in config file)
# {additional_host} - Extra host to send our notification to (from additional_host= in config file)
# {cache_file} - Cache file to write failed alarm sends to (from cache_file= in config file)
# {cache_expiration} - Seconds until alarms are considered too old to send (from cache_expiration= in config file)
# {logging} - Whether logging should occur (from logging= in config file) (1 = true | 0 = false)
# {log_file} - Log file to write to (from log_file= in config file)
# {debug} - Logging includes debugging (from debug= in config file or --debug on commandline) (1 = true | 0 = false)
if(! &parse_cmdline()) {
	&help(1);
}

# $galarms will be shared among subs and contain all of our alarm data for all hosts.
my $galarms={};
########################## $galarms data format ##################################
# {hosta}->{new}=[] - Array reference of hashes containing alarm data parsed from STDIN for sending to hosta.
#	 ->{old}=[] - Array reference of hashes containing alarm data from our cache file.
# {hostb}->{new}=[] - Same as above except alarms will be sent to 'hostb'
#	 ->{old}=[]
#
# Alarm Hash Format (contained in the arrays above).
# {hostname} - Hostname for the alarm.
# {servicename} - Service name for the alarm.  For host alarms this will be of zero length but
#		  still defined.
# {state} - Nagios numeric state of the alarm (0, 1, 2, 3)
# {output} - Service output.  Newlines should be escaped (literally \n, and not a newline).

&main();

sub main() {

	my $me="main";

	# Put a banner in our logs
	&w("debug", "----- Begin Execution -----");

	# Dump our configuration values into our log
	&w("debug", "$me: Configuration values:");
	foreach (keys %$gconf) {
		&w("debug", " - '$_' = '$gconf->{$_}'");
	}
	# See if we have a cache file defined.
	if(defined($gconf->{'cache_file'})) {
		# We have a cache_file defined, lets see if the file can be read.
		if(-r $gconf->{'cache_file'}) {
			&w("debug", "$me: A cache file exists at location '$gconf->{cache_file}'");
			# Read the cache file contents ($galarms is populated).
			if(! &parse_cache()) {
				&w("error", "There was a problem parsing the cache file at location '$gconf->{cache_file}'");
			} else {
				&w("debug", "$me Contents of the cache file was parsed into our alarms datatype");
			}
		}
	}
	
	# Read our new alarms from stdin ($galarms is populated).
	&w("debug", "$me: Attemping to read our alarms from STDIN");
	if(! &parse_stdin()) {
		&w("error", "There was a problem parsing alarms from STDIN.");
		if($gconf->{'parse_fatal'}) {
			&w("error", "parse_fatal is set to true. Errors parsing STDIN are considered fatal.  Exiting.");
			exit(2);
		}
	} else {
		&w("debug", "$me: Successfully parsed alarms from STDIN into our alarms datatype");
	}

	# Iterate over each host/alarm galarms()
	my $ec=0; my $send_count=0; my $fail_count=0;
	my ($host, $age, $xml);
	&w("debug", "$me: Iterating over alarm data, and sending alarms");
	foreach $host (keys %$galarms) {
		# Each host will have either an 'old' or 'new' key (or both)
		foreach $age (( 'old', 'new')) {
			# Old will be sent first, followed by new.  This should preseve any 'failures' then 'recoveries'
			if(defined $galarms->{$host}->{$age}) {
				&w("debug", "$me: Attempting to send '$age' alarms for host '$host'");
				# Note that we're sending something for our test later.
				# Build our XML file for this alarm set.
				$xml=&build_xml($galarms->{$host}->{$age});
				# If debugging is enabled we'll dump the XML we will send to our log.
				&w("debug", "$me: XML content for sending '$age' alarms to host '$host'");
				&w("debug", "$me: $xml");
				# send our alarms over.
				if(&send_alarm_http($host, $xml)) {
					$send_count+=scalar(@{$galarms->{$host}->{$age}});
					# We had a successful send, remove the record from the hash
					# Unsuccessful send will keep the entry in the hash, and we will store it in
					# cache for a later retry.
					&w("debug", "$me: Succeded sending $age alarms for host '$host'");
					delete($galarms->{$host}->{$age});
				} else {
					$fail_count+=scalar(@{$galarms->{$host}->{$age}});
					&w("error", "-- Failed to send '$age' alarms for host '$host' --");
					# we want to exit non-zero if any of our sends were not successful
					$ec=2;
				}
			}
		}
	}

	# After parsing above, any alarms that weren't forwarded properly will remain in our
	# galarms hash.  We can save them off for a retry at a later time if the cache file
	# is defined
	if(defined $gconf->{cache_file}) {
		# Save our alarms (if applicable).
		&w("debug", "$me: A cache file is defined in our configuration.  Will save any unsent alarms");
		&save_cache();
	}

	# If it turns out not valid alarms were found from STDIN nor from the cache file, we'll post an error.
	if(! $send_count) {
		&w("error", "$me: After parsing, no alarms were available for sending.");
		$ec=2;
	}

	&w("info", "Successfully delivered '$send_count' alarm(s) .  Failed to deliver '$fail_count' alarm(s)");
	
	&w("debug", "----- Finished execution.	Exiting error number '$ec' -----");

	exit($ec);
}

# - Read alarms from STDIN.
# - Place them into our galarms hash for each destination host under the {new} hash key.
sub parse_stdin() {

	my $me="parse_stdin";
	&w("debug", "$me: Called");

	my (@alarms, @split);
	my $ret=1;
	foreach my $line (<STDIN>) {
		chomp($line);
		&w("debug", "$me: Reading line '$line' from STDIN");
		# We will ignore empty, or lines containing nothing but spaces in STDIN.
		# Probably a stray newline, and we don't want to interfere with successful alerting.
		if($line=~m/^\s+$/ || $line=~m/^$/) {
			&w("debug", "$me: Found line '$line' was empty.  Skipped!");
			next;
		}
		my $raw={};
		# Split stdin on our delimiter.
		@split=split($gconf->{delim}, $line);

		# If old timestamps are specified on the commandline (-od) we expect
		# 4 elements for a host alarm, and 5 for a service alarm.
		if($gconf->{old_times}) {
			if(scalar(@split) == 4) {
				# HOST Alarm.
				$raw->{timestamp}=$split[0];
				$raw->{hostname}=$split[1];
				$raw->{servicename}="";
				$raw->{state}=$split[2];
				$raw->{output}=$split[3];
			} elsif(scalar(@split) == 5) {
				# SERVICE Alarm
				$raw->{timestamp}=$split[0];
				$raw->{hostname}=$split[1];
				$raw->{servicename}=$split[2];
				$raw->{state}=$split[3];
				$raw->{output}=$split[4];
			} else {
				&w("error", "The following line does not have enough elements to fire an alarm:");
				&w("error", "'$line'");
				&w("error", "Alarm not sent");
				$ret=0;
				next;
			}
		} else {
			# Old timestamps were not specified, so we'll use our current time.
			# A host alarm will contain 3 elemenets, and a service alarm 4
			if(scalar(@split) == 3) {
				# HOST Alarm
				$raw->{timestamp}=time();
				$raw->{hostname}=$split[0];
				$raw->{servicename}="";
				$raw->{state}=$split[1];
				$raw->{output}=$split[2];
			} elsif(scalar(@split) == 4) {
				$raw->{timestamp}=time();
				$raw->{hostname}=$split[0];
				$raw->{servicename}=$split[1];
				$raw->{state}=$split[2];
				$raw->{output}=$split[3];
			} else {
				&w("error", "The following line does not have enough elements to fire an alarm:");
				&w("error", "'$line'");
				&w("error", "Alarm not sent");
				$ret=0;
				next;
			}
		}
		# Our state needs to be 0 1 2 or 3.
		if($raw->{state}!~m/^(0|1|2|3)$/) {
			&w("error", "The following line had an invalid Nagios state. Must be 0, 1, 2, or 3:");
			&w("error", "'$line'");
			&w("error", "Alarm not sent");
			$ret=0;
			next;
		}
		&w("debug", "$me: Alarm content added to alarms array:");
		foreach (keys %$raw) {
			&w("debug", "$me:  '$_' = '$raw->{$_}");
		}
		push(@alarms, $raw);
	}

	# If we find no new alarms, we will return.  The {new} key won't be populated for any of our hosts.
	# Since its possible our cache file has pending alarms, we won't consider this fatal.
	if(scalar(@alarms) < 1) {
		&w("debug", "$me: After parsing STDIN, no valid alarms were found to send");
		&w("debug", "$me: Returning False");
		return(0);
	} else {
		&w("debug", "$me: After parsing STDIN, ".scalar(@alarms)." alarm(s) found to send");
	}

	# stdin alarms will be sent to our primary host (-H) from the commandline.
	# If there is no entry in galarms for this hostname create one.
	if(! defined $galarms->{$gconf->{host}}) {
		$galarms->{$gconf->{host}}={};
	}
	# Attach our array of hashes to this hosts 'new' key.
	$galarms->{$gconf->{host}}->{'new'}=\@alarms;

	# If our 'additional_host' is defined, we will add it as well.
	if(defined $gconf->{'additional_host'}) {
		if(! defined $galarms->{$gconf->{'additional_host'}}) {
			$galarms->{$gconf->{'additional_host'}}={};
		}
		$galarms->{$gconf->{'additional_host'}}->{'new'}=\@alarms;
	}

	&w("debug", "$me: Completed adding alarms to alarms datatype.");
	&w("debug", "$me: Returning True");
	return($ret);
}

# - Reads our cache file into our galarms hash into the {old} key.
# - The cache file has the following format:
#timestamp=1300220526
#failedhost=ulvgwnct1
#{
#	 hostname=ulvgwjst1
#	 servicename=gw-inf-svc-check_multiprocs
#	 state=0
#	 output=Everyuthing is very super|perf data\nLongoutput 1\nLongoutput2\nLongoutput3\n
#}
#timestamp=1300221024
#failedhost=ulvgwnct1
#{
#	 hostname=ulvgwjst1
#	 servicename=gw-inf-svc-check_multiprocs
#	 state=0
#	 output=Everyuthing is very super|perf data\nLongoutput 1\nLongoutput2\nLongoutput3\n
#}
# - The cache file is auto-generated, and shouldn't be messed with!
sub parse_cache() {

	my $me="parse_cache";
	&w("debug", "$me: Called");

	my @alarms;

	# If we've been called a '-r' test has already succeeded against the {cache_file} location
	# So if we fail to get a read handle, we consider that a problem.
	if(! open(FH, "<", $gconf->{cache_file})) {
		&w("error", "Could not get a read handle to cache file at location '$gconf->{cache_file}'");
		&w("error", $!);
		&w("debug", "$me: Returning False");
		return(0);
	} else {
		&w("debug", "$me: Successfully opened read handle to file '$gconf->{cache_file}'");
	}

	# Parse our data using our 'timestamp' and 'failedhost' entries that occur
	# prior to the open { and close } braces.  The file is auto-generated so we can
	# be relaxed with our expectations around comments etc.  No one should be messing with
	# the file.
	my $open=0; my $success; my $raw={};
	foreach my $line (<FH>) {
		chomp($line);

		&w("debug", "$me: Reading line '$line'");

		if($line=~m/^timestamp=(\d+)/) {
			$raw->{timestamp}=$1;
			&w("debug", "$me: Found 'timestamp' value of '$raw->{timestamp}'");
		}
		if($line=~m/^failedhost=(.*)/) {
			$raw->{failedhost}=$1;
			&w("debug", "$me: Found 'failedhost' value of '$raw->{failedhost}'");
		}
		# An open brace was found, set open to true.
		if($line=~m/^{/) {
			$open=1;
			&w("debug", "$me: Found an open brace");
		}

		# We're in an open brace, so we'll read name=value pairs into our raw reference.
		if($open) {
			if($line=~m/\s+(.*?)=(.*)/) {
				$raw->{$1}=$2;
				&w("debug", "$me: Found name value pair of '$1' = '$2'");
			}
		}

		# We've hit a close brace, validate our data held in $raw, and copy it off to the galarms if its valid.
		if($line=~m/^}/) {
			&w("debug", "$me: Found a close bracket");
			# make sure we have enough elements for this to be an alarm.
			$success=1;
			foreach my $test (( "timestamp", "failedhost", "hostname", "servicename", "state", "output")) {
				if(! defined $raw->{$test}) {
					# we're missing some data elements to consider this a valid alarm.  We can't use it.
					&w("error", "Failed to parse a result from the cache file. Possibly corrupt?");
					$success=0;
				}
			}
			if($success) {
				&w("debug", "$me: All alarm name=value pairs found");
				# copy our raw data into a new hash reference to store on an array.  We can't simply
				# push the raw reference onto the array or it will directly modify that reference next
				# iteration and we'll lose event data.
				my $newref={};
				foreach my $key (keys %$raw) {
					$newref->{$key}=$raw->{$key};
				}
				# see if we should expire our old alarms after a period of time.
				if(defined $gconf->{'cache_expiration'}) {
					# If our current time, minus the timestamp is greater than our expiration we
					# should discard this alarm.
					if((time() - $newref->{'timestamp'}) > $gconf->{'cache_expiration'}) {
						&w("debug", "$me: The following alarm was expired from cache due to age:");
						foreach my $key (keys %$newref) {
							&w("debug", "$me: - $key=$newref->{$key}");
						}
					} else {
						&w("debug", "$me: Cache expiration is enabled, but does not apply to this alarm");
						# alarm is still current.
						push(@alarms, $newref);
					}
				} else {
					push(@alarms, $newref);
				}
			}
			# Set our read back to 'closed'.
			$open=0;
			# Empty out our raw hash for the next iteration.
			$raw={};
		}
	}
	&w("debug", "$me: After parsing cache file, found ".scalar(@alarms)." to retry");
	&w("debug", "$me: Completed parsing cache file, closing read handle");
	close(FH);

	# Now we can go over our alarms array of hashes and pack them onto the right spot in our galarms.
	# They will be placed in {host}->{old} where {host} correlates with {failedhost} from our cache file.
	&w("debug", "$me: Placing raw alarm data into our alarm datatype");
	foreach my $hash (@alarms) {
		# make sure an entry exists for this host in our main galarms hash
		if(! defined $galarms->{$hash->{failedhost}}) {
			$galarms->{$hash->{failedhost}}={};
			$galarms->{$hash->{failedhost}}->{'old'}=[];
		}
		# see if we have an array reference in our old key
		if(ref $galarms->{$hash->{failedhost}}->{'old'} ne "ARRAY") {
			# create an anonymous hash in the 'old' key
			$galarms->{$hash->{failedhost}}->{'old'}=[];
		}
		# push our hash onto the array reference.
		push(@{$galarms->{$hash->{'failedhost'}}->{'old'}}, $hash);
	}

	&w("debug", "$me: Returning True");
	return(1);
}

# - Save off our cache file to disk.
# - See 'parse_cache' for cache file formatting.
sub save_cache() {

	my $me="save_cache";
	&w("debug", "$me: Called");

	# First we should determine if there are any alarms left in our alarms global hash.
	# If there isn't we can return success.
	my $write=0;
	foreach my $host (keys %$galarms) {
		foreach my $age (( 'old', 'new' )) {
			if(defined $galarms->{$host}->{$age}) {
				foreach my $hash (@{$galarms->{$host}->{$age}}) {
					$write=1;
					last;
				}
			}
		}
	}

	# See if we have some data to write.
	if(! $write) {
		&w("debug", "$me: Determined there were no unsent alarms to write to cache");
		if(-r $gconf->{cache_file}) {
			if(unlink($gconf->{cache_file})) {
				&w("debug", "$me: Found a cache file existed, removed from disk");
				&w("debug", "$me: Return True");
				return(1);
			} else {
				&w("error", "A cache file exists but should be removed!");
				&w("error", "Failed to remove file $gconf->{cache_file}: $!");
				&w("debug", "$me: Return False");
				return(0);
			}
		}
	}

	# Unsent alarms exist that we need to cache for next execution.
	if(! open(FH, ">", $gconf->{cache_file})) {
		&w("error", "Failed to get a write handle to cache file at location '$gconf->{cache_file}'");
		&w("error", $!);
		&w("debug", "$me: Return False");
		return(0);
	}

	# Print a header to warn against modification.
	&w("debug", "$me: Writing a header to our cache file");
	print FH "# Auto-generated by Groundwork send_nrdp.pl.	Do not directly edit!\n";
	print FH "# Direct any questions to the ESM group\n";

	# Write will be set to 1 if we have a new alarm lines.
	# This is easier then testing the {old} {new} key lengths etc for each host to decide
	# Whether all sends were successful.
	&w("debug", "$me: Iterating over unsent alarm content in alarm global hash");
	my $count=0;
	foreach my $host (keys %$galarms) {
		foreach my $age (( 'old', 'new' )) {
			if(defined $galarms->{$host}->{$age}) {
				foreach my $hash (@{$galarms->{$host}->{$age}}) {
					$count++;
					print FH "timestamp=$hash->{timestamp}\n";
					print FH "failedhost=$host\n";
					print FH "{\n";
					print FH "\thostname=$hash->{hostname}\n";
					print FH "\tservicename=$hash->{servicename}\n";
					print FH "\tstate=$hash->{state}\n";
					print FH "\toutput=$hash->{output}\n";
					print FH "}\n";
					&w("debug", "$me: Caching future alarms for recipient host '$host':");
					foreach (('timestamp', 'hostname', 'servicename', 'state', 'output')) {
						&w("debug", "$me:  '$_' = '$hash->{$_}'");
					}
				}
			}
		}
	}

	close(FH);

	&w("debug", "$me: Closing write filehandle");
	&w("debug", "$me: A total of '$count' alarms were written to cache");
	&w("debug", "$me: Return True");

	return(1);
}

# - Creates the XML for sending to the listening send_nrdp URL.
# - Sorts alarm data based on its timestamp so any older alarms arrive at the host first
# - Escapes all XML special characters.
# IN: Array reference containing hash references for our alarm data
# OUT: Scalar containing the XML document to post to the host.
sub build_xml() {

	my $alarms=shift;

	my $me="build_xml";
	&w("debug", "$me: Called with ".scalar(@$alarms)." alarms to format into XML");

	# We need to sort the alarms we're working with by their timestamp so any
	# older alarms we're sending respect order.  We might have some criticals followed by OKs etc
	# and if our order is wrong it can give us an unexpected state.

	# first find out how many unique timestamps we have in our array of alarm hashes.
	&w("debug", "$me: Sorting alarms by their timestamp value");
	my $timestamps={}; my $alarm; my $timestamp;
	foreach $alarm (@$alarms) {
		$timestamps->{$alarm->{'timestamp'}}={};
	}

	# Now iterate over each timestamp, and we'll build a sorted_alarms array.
	my @sorted_alarms;
	foreach $timestamp (sort(keys %$timestamps)) {
		&w("debug", "$me: Sorting alarms with timestamp '$timestamp'");
		foreach $alarm (@$alarms) {
			if($alarm->{'timestamp'} == $timestamp) {
				push(@sorted_alarms, $alarm);
			}
		}
	}

	# Our XML header will be consistent.
	my $xml=qq(<?xml version='1.0'?>\n);
	$xml.=qq(<checkresults>\n);

	# Now we can add our 'checkresult' data for each alarm in our array.
	foreach $alarm (@sorted_alarms) {
		# If servicename has a zero length then this is a host type alarm.
		if(length($alarm->{servicename}) < 1) {
			$xml.=qq(\t<checkresult type='host' checktype='1'>\n);
		} else {
			$xml.=qq(\t<checkresult type='service' checktype='1'>\n);
		}
		$xml.="\t\t<hostname>".&escape_xml($alarm->{hostname})."</hostname>\n";
		if(length($alarm->{servicename}) > 0) {
			$xml.="\t\t<servicename>".&escape_xml($alarm->{servicename})."</servicename>\n";
		}
		$xml.=qq(\t\t<state>$alarm->{state}</state>\n);
		$xml.="\t\t<output>".&escape_xml($alarm->{output})."</output>\n";
		$xml.=qq(\t</checkresult>\n);
	}
	# Button up our XML with our close checkresults brace.
	$xml.=qq(</checkresults>\n);

	&w("debug", "Returning XML results");

	return($xml);
}

# - Make a connection to port 80 on a remote host.
# - POST a url encounted form to the URL provided for NRDP.
# - Read the results from the POST and look for he <message>OK</message> return
# IN: SCalar containg the hostname to contact.
# IN: Scalar containing the XML to post.
# OUT: 1 for success, 0 for failure.
sub send_alarm_http() {


	my $host=shift;
	my $xml=shift;

	my $me="send_alarm_http";
	&w("debug", "$me: Called for host '$host'");

	# Try to open our socket
	&w("debug", "$me: Attempting to open socket to port 80 via tcp on host '$host' with timeout '$gconf->{timeout}'");
	my $sock=IO::Socket::INET->new(	PeerAddr	=>	$host,
					PeerPort	=>	'http(80)',
					Proto		=>	'tcp',
					Timeout		=>	$gconf->{'timeout'});
	if(! $sock) {
		&w("error", "Failed to open a socket to port 80 on host '$host'");
		&w("error", "Error: $!");
		&w("debug", "$me: Return False");
		return(0);
	}

	# Escape our XML into URL formatting.
	my $escaped_xml=&escape_url($xml);

	# Our URL will contain our token, our command (submitcheck) and our URLized XML data.
	my $url="token=$gconf->{token}&cmd=submitcheck&XMLDATA=$escaped_xml";
	&w("debug", "$me: Final URL to post is:");
	&w("debug", "$me: $url");


	# Our Post document:
	# (I made up the 'Send NRDP Perl Agent'.  No idea if that violates some RFC or something.  If it does, sorry).
	my $post;
	$post.="POST $gconf->{url} HTTP/1.1\r\n";
	$post.="Host: $host\r\n";
	$post.="Accept: text/*\r\n";
	$post.="User-Agent: send_nrdp.pl/$version/Send NRDP Perl Agent\r\n";
	$post.="Connection: Close\r\n";
	# If we have a use rand password specified we will encode for basic autheentication.
	if(defined $gconf->{'http_user'}) {	
		my $user_pass=&encode_base64(sprintf("%s:%s", $gconf->{'http_user'}, $gconf->{'http_pass'}));
		# For some reason encode_base64 is adding a newline, chomp it off.
		chomp($user_pass);
		$post.="Authorization: Basic $user_pass\r\n";
	}
	$post.="Content-Type: application/x-www-form-urlencoded\r\n";
	$post.="Content-Length: ".length($url)."\r\n";
	$post.="\r\n$url";

	&w("debug", "$me: Final POST document to send is:");
	&w("debug", "$me: $post");

	# Place our data in the socket.
	print $sock $post;
	# Read our response into the @out array.
	my @out=<$sock>;
	# Close our connection.
	close($sock);

	# See if we got a successful message or not.
	my $success=0;
	foreach(@out) {
		if(m/^\s+\<message\>OK\<\/message\>$/) {
			$success=1;
		}
	}

	# if it worked, we'll post our return as debugging data.
	if($success) {
		&w("debug", "$me: Host: '$host'. Package accepted for delivery.  HTTP response follows:");
		foreach(@out) {
			chomp();
			&w("debug", "  $_");
		}
		&w("debug", "Return True");
		return(1);
	} else {
	# IF it failed our output will be treated as error data.
		&w("error", "Host: '$host'. Package not accepted for delivery.	HTTP response follows:");
		foreach(@out) {
			chomp();
			&w("error", "  $_");
		}
		&w("debug", "Return False");
		return(0);
	}
}

# - Reads our commandline, and optinoally our configuration file
#   into the gconf global hash (see header for format).
# - Validate as much as possible to make sure execution will work.
# - Provide defaults for unspecified values that are requird for operation.
sub parse_cmdline() {

	# Parse our raw commandline data into our $c hash references.
	my $c={};
	GetOptions($c, 'H=s', 'p=i', 'd=s', 'to=i', 'c=s', 'url=s', 'token=s', 'http_user=s', 'http_pass=s', 'debug');

	################## Parse values.

	# Since this can be used as a plugin replacement for send_nsca we wouldn't normally get a URL
	# so the URL will either be embedded in the configuration file, or passed on the commandline.
	# First thing we'll do is see if we have a config file.  If so, we'll parse it in.
	if(defined $c->{'c'}) {
		# Make sure we can read the file.
		if(! -r $c->{'c'}) {
			print "\nConfig file was specified (-c).\n";
			print "But could not be read at location '$c->{c}'\n";
			return(0);
		}
		if(! &parse_config($c->{'c'})) {
			print "\nConfig file was specified (-c).\n";
			print "Encountered errors while parsing file '$c->{c}'\n";
			return(0);
		}
	}

	# Assign our hostname if it was defined.
	if(defined $c->{'H'}) {
		$gconf->{'host'}=$c->{'H'};
	}

	# See if we got a delimiter specified.
	if(defined $c->{'d'}) {
		$gconf->{'delim'}=$c->{'d'};
	} else {
		# We use our default.
		$gconf->{'delim'}="\t";
	}

	# See if we had a timeout specified.	
	if(defined $c->{'to'}) {
		$gconf->{'timeout'}=$c->{'to'};
	} else {
		$gconf->{'timeout'}=10;
	}

	# See if we were told to use old timestamps.
	if(defined $c->{'od'}) {
		$gconf->{'old_times'}=1;
	} else {
		$gconf->{'old_times'}=0;
	}

	# Values on the commandline can override ones from the config file.

	# See if we got a url specified
	if(defined $c->{'url'}) {
		# A value exists on the commandline, we'll use it.
		$gconf->{'url'}=$c->{'url'};
	}

	# See if we had a token on the commandline.
	if(defined $c->{'token'}) {
		$gconf->{'token'}=$c->{'token'};
	}

	# See if we had a user on the commandline.
	if(defined $c->{'http_user'}) {
		$gconf->{'http_user'}=$c->{'http_user'};
	}

	# See if we had a password on the commandline.
	if(defined $c->{'http_pass'}) {
		$gconf->{'http_pass'}=$c->{'http_pass'};
	}

	# See if debugging was specified.
	if(defined $c->{'debug'}) {
		$gconf->{'debug'}=$c->{'debug'};
	} else {
		# We may have no value for debug from either the CLI or the conf file.
		if(! defined $gconf->{'debug'}) {
			$gconf->{'debug'}=0;
		}
	}

	# We may have no value in the conf file for parse_fatal.  Our default will be 1
	if(! defined $gconf->{parse_fatal}) {
		$gconf->{parse_fatal}=1;
	}

	################## Validate values.

	my $success=1;
	# URL must be specified
	if(! defined $gconf->{'url'}) {
		print "\nA URL to send results to was not specified either on the commandline (--url)\n";
		print "or via the configuration file.  Cannot send results.\n";
		$success=0;
	}

	# A host to send results to must be specified.
	if(! defined $gconf->{'host'}) {
		print "\nA hostname to send results to was not specified (-H)\n";
		print "Cannot send results.\n";
		$success=0;
	}

	# A token must be specified.
	if(! defined $gconf->{'token'}) {
		print "\nA Token to send results was not specified either on the commandline (--token)\n";
		print "or via the configuration file.  Cannot send results.\n";
		$success=0;
	}

	# if 'debug' or 'logging' was set, we must have a numeric value.
	foreach my $check (( 'debug', 'logging', 'parse_fatal')) {
		if(defined $gconf->{$check} ) {
			if($gconf->{$check}!~m/^\d+$/) {
				print "\nValue for '$check' must be numeric only.  Found value '$gconf->{$check}'.  Cannot continue.\n";
				$success=0;
			}
		}
	}

	# If 'logging' is set, we must have a 'log_file' value as well.
	# If we don't have it, it isn't conidered fatal, but we won't be able to log.
	if(defined $gconf->{'logging'}) {
		if($gconf->{'logging'} && ! defined $gconf->{'log_file'}) {
			print "\n'logging' is set to true, but no 'log_file' was specified in the configuration file.\n";
			print "\nLogging will not be possible.\n";
			$gconf->{logging}=0;
		}
	}

	# If an http user is specified, we must also have a password.
	if(defined $gconf->{'http_user'}) {
		if(! defined $gconf->{'http_pass'}) {
			print "\n(-http_user) was specified, http password (-http_pass) is also required\n";
			$success=0;
		}
	}

	return($success);
}

# - Reads our configuration file looking for very specific name=value pairs.
# - Since we could have numerous other name=value pairs for send_nsca config files we're selective.
sub parse_config() {

	my $file=shift;

	# We've done a -r test before calling parse_config, so failing to open a handle is a problem.
	if(! open(FH, "<", $file)) {
		print "\nUnable to open read handle to configuration file '$file'.  Cannot continue\n";
		print "Error: $!\n";
		return(0);
	}

	# Read all our name=value pairs into a raw hash.
	my $raw={};
	while(<FH>) {
		if(m/^#\s?NRDP\s?(.*?)\s?=\s?(.*)$/) {
			$raw->{$1}=$2;
		}
	}
	close(FH);

	# Remove the specific send_nrdp.pl name=value pairs from our $raw hash.
	foreach my $keyword ((	'additional_host', 'cache_file', 'cache_expiration', 'url', 'token', 'http_user', 'http_pass', 'logging', 
				'debug', 'log_file', 'parse_fatal' )) {
		if(defined $raw->{$keyword}) {
			$gconf->{$keyword}=$raw->{$keyword};
		}
	}
	return(1);
}

# - Our general 'write' function to either write to stdout/stderr or a log file deending on configuration.
# - If logging is enabled, we will write to a log file.
# - If logging isn't enabled, we'll log to STDOUT/STDERR
sub w {

	my $type=shift;
	my $log_data=shift;

	# if debug isn't enabled, and its a debug message, then return without action.
	if($type eq "debug" && $gconf->{'debug'} < 1) {
		return(1);
	}

	# Otherwise we should be good to write data to our file.
	my @time=localtime();
	# MM-DD-YY HH:MM:SS
	my $timestamp=sprintf("%02d-%02d-%2d %02d:%02d:%02d", $time[4]+1, $time[3], $time[5]-100, $time[2], $time[1], $time[0]);

	# translate our type ot all caps.  Gets more attention, and I didn't like typing in caps when coding :)
	$type=~tr/[a-z]/[A-Z/;

	# If logging is enabled, we'll write our information to our log file.
	if($gconf->{'logging'}) {
		if(! open(LFH, ">>", $gconf->{'log_file'})) {
			print STDERR "Unable to open log file $gconf->{log_file} for writing!\n";
			return(0);
		}

		# see if we have a module name set, by default we won't.
		print LFH "$timestamp - $type - $log_data\n";

		close(LFH);
	} else {
	# Logging isn't enabled, so we'll write our data to STDERR for 'error' severity,
	# or STDOUT for the rest (info, and debug).
		if($type eq "ERROR") {
			print STDERR "$timestamp - $type - $log_data\n";
		} else {
			print STDOUT "$timestamp - $type - $log_data\n";
		}
	}

	return(1);
}

# Translates XML special characters to their XML escaped equivelents.
# IN: Scalar containing text data to escape
# OUT: Scalar containing Escaped data
sub escape_xml() {

	my $xml=shift;

	# First we'll remove all characters are are invalid in XML 1.0.
	$xml=~s/[^\x09\x0A\x0D\x20-\x{D7FF}\x{E000}-\x{FFFD}\x{10000}-\x{10FFFF}]//go;

	# Ampersand needs to be at the top of this list, or else the XML ampersand will get escaped
	# remaining order doesn't matter.
	$xml=~s/&/\&amp;/g;
	$xml=~s/'/\&apos;/g;
	$xml=~s/"/\&quot;/g;
	$xml=~s/</\&lt;/g;
	$xml=~s/>/\&gt;/g;

	return($xml);
}

# Translates textual characters to their URL equivilents.
# (Yes there are CPAN modules that do this, but I'm trying to keep this non-CPAN for expaneded use).
# IN: Scalar containing text data to escape
# OUT: Scalar containing Escaped data
sub escape_url() {

	my $url=shift;

	# Got this off the internets.  Seems to work well!	
	$url =~ s/([^A-Za-z0-9])/sprintf("%%%02X", ord($1))/seg;

	return($url);
}

# Lifted directly from the MIME::Base64 encoding CPAN module at:
# http://cpansearch.perl.org/src/GAAS/MIME-Base64-Perl-1.00/lib/MIME/Base64/Perl.pm
# - Some of our base Unix platforms don't include this CPAN module in their
#   distribution and we need to base64 encode the user and password for basic authentication
sub encode_base64 ($;$)
{
	if ($] >= 5.006) {
		require bytes;
		if (bytes::length($_[0]) > length($_[0]) ||
			($] >= 5.008 && $_[0] =~ /[^\0-\xFF]/)) {
			    require Carp;
			    Carp::croak("The Base64 encoding is only defined for bytes");
		}
	}

	use integer;

	my $eol = $_[1];
	$eol = "\n" unless defined $eol;

	my $res = pack("u", $_[0]);
	# Remove first character of each line, remove newlines
	$res =~ s/^.//mg;
	$res =~ s/\n//g;

	$res =~ tr|` -_|AA-Za-z0-9+/|;		 # `# help emacs
	# fix padding at the end
	my $padding = (3 - length($_[0]) % 3) % 3;
	$res =~ s/.{$padding}$/'=' x $padding/e if $padding;
	# break encoded string into lines of no more than 76 characters each
	if (length $eol) {
		$res =~ s/(.{1,76})/$1$eol/g;
	}
	return $res;
}

# - Prints our help, and exits with the value we've been called with.
sub help() {

	my $ec=shift;

	print "\nNRDP Perl Client Version $version\n";
###############################################################################
	print <<EOF
Usage: send_nrdp.pl -H <host_address> [-p port] [-to to_sec] [-d delim] 
		    [-c config_file] [-od] [-url url] [-token token]
		    [-http_user user] [-http_pass pass] --debug

Options:
 <host_address> = The IP address of the host with the NRDP URL
 [port]		= Port Number to use (**)
 [to_sec]	= Number of seconds before connection attempt times out.
		  (default timeout is 10 seconds)
 [delim]	= Delimiter to use when parsing input (defaults to a tab)
 [config_file]	= Name of config file to use
 [-od]		= Specify old timestamp (**)
 [url]		= URL that the NRDP CGI is available on
 [token]	= Token to use for NRDP submission
 [user]		= Username for HTTP basic authentication (not secured)
 [pass]		= Password for HTTP basic authentication (not secured)
 [--debug]	= Print debug messages

NOTE: Options marked (**) exist for backward compatibility with send_nsca, but
      do not impact the operation of send_nrdp.pl.

This utility is used to send passive check results to the NRDP CGI.  Host and
Service check data that is to be sent to the NRDP CGI is read from standard
input. Input should be provided in the following format (tab-delimited unless
overriden with -d command line argument, one entry per line):

Service Checks:
<host_name>[tab]<svc_description>[tab]<return_code>[tab]<plugin_output>[newline]

Host Checks:
<host_name>[tab]<return_code>[tab]<plugin_output>[newline]

NOTE: NRDP specific options (-url, -token, -u, -p, --debug) can be specified in 
      the configuration file rather than command line to maintain compatability 
      with existing send_nsca scripts.	Also, there are additional items that can
      be specified in the config file to control execution.  See documentation.

EOF
;
	exit($ec);
}
